# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# Segment.py
# Python implementation of the Class Segment
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:11:57 AM
# Original author: talve
#
#######################################################
from abc import ABC, abstractmethod
from collections import OrderedDict


class Segment(ABC):
    """this class is responsible for holding segment data according to its type.
    """
    def __init__(self):
        """initialize the class by setting the class data.
        """
        self.size = 0
        self._parsed_data = OrderedDict()

    def additional_title_info(self):
        """return index1 and index2 if exists in the segment.
        """
        return ""

    def get_size(self):
        return self.size

    @abstractmethod
    def get_data(self):
        """get the segment data.
        """
        pass

    @abstractmethod
    def get_type(self):
        """get the segment type.
        """
        pass

    def get_data_in_bytes(self):
        """This method get the segment data in a list of bytes
        include the header.
        """
        seg_bytes = []
        for dw in self.get_data():
            for byte in dw.to_bytes(4, byteorder='big'):
                seg_bytes.append(byte)
        return seg_bytes

    def get_parsed_data(self):
        """get dictionary of parsed segment data.
        """
        return []

    def add_parsed_data(self, key, value):
        self._parsed_data[key] = value

    def get_parsed_data(self):
        """get dictionary of parsed segment data.
        """
        return self._parsed_data



# Copyright (c) 2004-2010 Mellanox Technologies LTD. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --

from __future__ import print_function
import os
import sys
import platform
import ctypes
import mtcr


class DevMgtException(Exception):
    pass


##########################
DEV_MGT = None
try:
    from ctypes import *
    if platform.system() == "Windows" or os.name == "nt":
        try:
            DEV_MGT = CDLL("c_dev_mgt.dll")
        except BaseException:
            DEV_MGT = CDLL(os.path.join(os.path.dirname(os.path.realpath(__file__)), "c_dev_mgt.dll"))
    else:
        try:
            # c_dev_mgt.so must be specified explicitly for pyinstaller collection
            DEV_MGT = CDLL("c_dev_mgt.so")
        except BaseException:
            DEV_MGT = CDLL(os.path.join(os.path.dirname(os.path.realpath(__file__)), "c_dev_mgt.so"))
except Exception as exp:
    raise DevMgtException("Failed to load shared library c_dev_mgt: %s" % exp)

if DEV_MGT:
    class DevMgt:
        ##########################
        def __init__(self, dev, i2c_secondary= -1):
            self.mf = 0
            self._isLivefishModeFunc = DEV_MGT.dm_is_livefish_mode  # dm_is_livefish_mode(mfile* mf)
            self._getDeviceIdFunc = DEV_MGT.dm_get_device_id
            self._isIBAccess = DEV_MGT.dm_is_ib_access
            if isinstance(dev, (mtcr.MstDevice)):
                self._mstdev = dev
            else:
                self._mstdev = mtcr.MstDevice(dev, i2c_secondary)
                if not self._mstdev:
                    raise DevMgtException("Failed to open device (%s): %s" % (dev, os.strerror(ctypes.get_errno())))

        ##########################
        def isLivefishMode(self):
            rc = self._isLivefishModeFunc(self._mstdev.mf)
            return not (rc == 0)

        ##########################
        def getHwId(self):
            dm_dev_id = ctypes.c_uint16()
            hw_dev_id = ctypes.c_uint32()
            hw_rev = ctypes.c_uint32()

            rc = DEV_MGT.dm_get_device_id(self._mstdev.mf, ctypes.byref(dm_dev_id), ctypes.byref(hw_dev_id), ctypes.byref(hw_rev))
            return {"return_code": rc, "dm_device_id": dm_dev_id, "hw_device_id": hw_dev_id, "hw_revision": hw_rev}

        @classmethod
        def getDevIdOffline(cls, hw_dev_id):
            dm_dev_id = ctypes.c_uint16()
            rc = DEV_MGT.dm_get_device_id_offline(hw_dev_id, -1, ctypes.byref(dm_dev_id))
            return {"return_code": rc, "dm_device_id": dm_dev_id, "hw_device_id": hw_dev_id, "hw_revision": -1}

        @classmethod
        def str2hwId(cls, name):
            name_char_p = ctypes.c_char_p(name.lower().encode('utf-8'))
            dm_dev_id = ctypes.c_uint16()
            hw_dev_id = ctypes.c_uint32()
            hw_rev = ctypes.c_uint32()

            dm_dev_id = DEV_MGT.dm_dev_aproxstr2type(name_char_p)

            rc = 0
            hw_dev_id = DEV_MGT.dm_get_hw_dev_id(dm_dev_id)
            hw_rev = DEV_MGT.dm_get_hw_rev_id(dm_dev_id)

            if (hw_dev_id == 0 or hw_rev == 0):
                rc = 1

            return {"return_code": rc, "dm_device_id": dm_dev_id, "hw_device_id": hw_dev_id, "hw_revision": hw_rev}

        @classmethod
        def type2str(cls, dm_dev_id):
            DEV_MGT.dm_dev_type2str.restype = ctypes.c_char_p
            return DEV_MGT.dm_dev_type2str(dm_dev_id).decode("utf-8")

        @classmethod
        def is_ib_switch(cls, dm_dev_id):
            return DEV_MGT.dm_dev_is_ib_switch(dm_dev_id)

        @classmethod
        def is_connectib(cls, dm_dev_id):
            return DEV_MGT.dm_is_connectib(dm_dev_id)

        @classmethod
        def is_gb100(cls, dm_dev_id):
            return DEV_MGT.dm_is_gb100(dm_dev_id)

        @classmethod
        def is_cx7(cls, dm_dev_id):
            return DEV_MGT.dm_is_cx7(dm_dev_id)

        @classmethod
        def is_fs3(cls, dm_dev_id):
            return DEV_MGT.dm_dev_is_fs3(dm_dev_id)

        @classmethod
        def is_fs4(cls, dm_dev_id):
            return DEV_MGT.dm_dev_is_fs4(dm_dev_id)

        @classmethod
        def is_4th_gen(cls, dm_dev_id):
            return DEV_MGT.dm_is_4th_gen(dm_dev_id)

        @classmethod
        def sw_id2hw_id(cls, sw_dev_id):
            dm_dev_id = DEV_MGT.dm_dev_sw_id2type(sw_dev_id)
            rc = 0
            hw_dev_id = DEV_MGT.dm_get_hw_dev_id(dm_dev_id)
            hw_rev = DEV_MGT.dm_get_hw_rev_id(dm_dev_id)

            if (hw_dev_id == 0 or hw_rev == 0):
                rc = 1

            return {"return_code": rc, "dm_device_id": dm_dev_id, "hw_device_id": hw_dev_id, "hw_revision": hw_rev}

        def is_ib_access(self):
            return (self._isIBAccess(self._mstdev.mf) != 0)
else:
    raise DevMgtException("Failed to load c_dev_mgt.so/libdev_mgt.dll")



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# constants.py
# Python implementation of resource dump constants
# This file will supply all the constants needed
# for the resource dump tool
# Original author: romany
#
#######################################################

TOOL_NAME = "ResourceDump"
RESOURCE_DUMP_COMMAND_TYPE_QUERY = "query"
RESOURCE_DUMP_COMMAND_TYPE_DUMP = "dump"

RESOURCE_DUMP_SEGMENT_TYPE_RESOURCE = "0xffffffff"
RESOURCE_DUMP_SEGMENT_TYPE_RESOURCE_MIN = "0x0000"
RESOURCE_DUMP_SEGMENT_TYPE_RESOURCE_MAX = "0xfeff"
RESOURCE_DUMP_SEGMENT_TYPE_NOTICE = "0xfff9"
RESOURCE_DUMP_SEGMENT_TYPE_COMMAND = "0xfffa"
RESOURCE_DUMP_SEGMENT_TYPE_TERMINATE = "0xfffb"
RESOURCE_DUMP_SEGMENT_TYPE_ERROR = "0xfffc"
RESOURCE_DUMP_SEGMENT_TYPE_INFO = "0xfffe"
RESOURCE_DUMP_SEGMENT_TYPE_MENU = "0xffff"
RESOURCE_DUMP_SEGMENT_TYPE_REFERENCE = "0xfffd"


UI_DASHES = "--"
UI_DASHES_SHORT = "-"
UI_ARG_DEVICE = "device"
UI_ARG_DEVICE_SHORT = "d"
UI_ARG_SEGMENT = "segment"
UI_ARG_SEGMENT_SHORT = "s"
UI_ARG_VHCAID = "virtual_hca_id"
UI_ARG_VHCAID_SHORT = "v"
UI_ARG_INDEX1 = "index1"
UI_ARG_INDEX1_SHORT = "i1"
UI_ARG_INDEX2 = "index2"
UI_ARG_INDEX2_SHORT = "i2"
UI_ARG_NUMOFOBJ1 = "num_of_obj1"
UI_ARG_NUMOFOBJ1_SHORT = "n1"
UI_ARG_NUMOFOBJ2 = "num_of_obj2"
UI_ARG_NUMOFOBJ2_SHORT = "n2"
UI_ARG_DEPTH = "depth"
UI_ARG_DEPTH_SHORT = "de"
UI_ARG_BIN = "bin"
UI_ARG_BIN_SHORT = "b"
UI_ARG_MEM = "mem"
UI_ARG_MEM_SHORT = "m"

# segment start/end specify where the 'type' attr starts and ends in the raw_data of each segment.
# should be 0-16 bits
SEGMENT_TYPE_DWORD_LOCATION = 0
SEGMENT_SIZE_DWORD_LOCATION = 0
SEGMENT_TYPE_START = 16
SEGMENT_TYPE_END = 32
SEGMENT_SIZE_START = 0
SEGMENT_SIZE_END = 16
MENU_SEGMENT_NUM_OF_RECORDS_DWORD_LOCATION = 1
MENU_SEGMENT_NUM_OF_RECORDS_START = 16
MENU_SEGMENT_NUM_OF_RECORDS_END = 32

MENU_SEGMENT_TYPE_DWORD_LOCATION = 0
MENU_SEGMENT_DWORD_SUPPORT_ATTRS_LOCATION = 0
MENU_SEGMENT_SEGMENT_NAME_DWORD_LOCATION_START = 1
MENU_SEGMENT_SEGMENT_NAME_DWORD_LOCATION_END = 5
MENU_SEGMENT_INDEX_NAME_1_DWORD_LOCATION_START = 5
MENU_SEGMENT_INDEX_NAME_1_DWORD_LOCATION_END = 9
MENU_SEGMENT_INDEX_NAME_2_DWORD_LOCATION_START = 9
MENU_SEGMENT_INDEX_NAME_2_DWORD_LOCATION_END = 13

NUM_OF_OBJ_ALL = 65535
NUM_OF_OBJ_ACTIVE = 65534

MENU_SEGMENT_TABLE_SEGMENT_TYPE = "Segment Type - "
MENU_SEGMENT_TABLE_DUMP_PARAMS = "Dump Params"
MENU_SEGMENT_TABLE_APPLICABILITY = "Applicability"
MENU_SEGMENT_TABLE_SPECIAL_VALUES = "Special Values"
MENU_SEGMENT_TABLE_MANDATORY = "Mandatory"
MENU_SEGMENT_TABLE_OPTIONAL = "Optional"
MENU_SEGMENT_TABLE_INDEX1 = UI_ARG_INDEX1
MENU_SEGMENT_TABLE_INDEX2 = UI_ARG_INDEX2
MENU_SEGMENT_TABLE_NUM_OF_OBJS1 = UI_ARG_NUMOFOBJ1
MENU_SEGMENT_TABLE_NUM_OF_OBJS2 = UI_ARG_NUMOFOBJ2
MENU_SEGMENT_TABLE_ALL = "all"
MENU_SEGMENT_TABLE_ACTIVE = "active"
MENU_SEGMENT_TABLE_LINE_LEN = 66
MENU_SEGMENT_TABLE_DUMP_PARAMS_LEN = 32
MENU_SEGMENT_TABLE_APPLICABILITY_LEN = 14
MENU_SEGMENT_TABLE_SPECIAL_VALUES_LEN = 14
MENU_SEGMENT_TABLE_GAP = 3
MENU_SEGMENT_TABLE_NA = "N/A"
MENU_SEGMENT_TABLE_DASH = "-"
MENU_SEGMENT_TABLE_AND = ","
MENU_SEGMENT_TABLE_WRAP = "_"

MKEY_BUF_SIZE = 262144  # 2^18 bytes



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# CommandSegment.py
# Python implementation of the Class CommandSegment
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:11:57 AM
# Original author: talve
#
#######################################################
from segments.Segment import Segment
from segments.SegmentFactory import SegmentFactory
from utils import constants as cs


class CommandSegment(Segment):
    """this class is responsible for holding command segment data.
    """

    def __init__(self, data):
        """initialize the class by setting the class data.
        """
        super().__init__()
        self.raw_data = data
        self.index1 = self.raw_data[2]
        self.index2 = self.raw_data[3]
    
    def additional_title_info(self):
        """return index1 and index2 if exists in the segment.
        """
        return " ; index1 = {0}, index2 = {1}".format(hex(self.index1), hex(self.index2))

    def get_data(self):
        """get the general segment data.
        """
        return self.raw_data

    def get_type(self):
        """get the general segment type.
        """
        return cs.RESOURCE_DUMP_SEGMENT_TYPE_COMMAND

    def get_parsed_data(self):
        """get dictionary of parsed segment data.
        """
        # dw_data = self.get_data()
        # self._parsed_data['Segment Type'] = "{0} ({1})".format("Command Segment",
        #                                                  str(hex(int('{:0b}'.format(dw_data[0]).zfill(32)[16:32], 2))))
        # offset = cs.SEGMENTS_HEADER_SIZE_IN_DW
        #
        # if len(dw_data) > cs.SEGMENTS_HEADER_SIZE_IN_DW:
        #     self._parsed_data['segment_called'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[0:16], 2)))
        #     self._parsed_data['vhca_id'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[16:32], 2)))
        #     offset += 1
        #     self._parsed_data['index1'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[0:32], 2)))
        #     offset += 1
        #     self._parsed_data['index2'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[0:32], 2)))
        #     offset += 1
        #     self._parsed_data['num_of_obj1'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[0:16], 2)))
        #     self._parsed_data['num_of_obj2'] = str(hex(int('{:0b}'.format(dw_data[offset]).zfill(32)[16:32], 2)))
        return self._parsed_data


SegmentFactory.register(cs.RESOURCE_DUMP_SEGMENT_TYPE_COMMAND, CommandSegment)



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# QueryData.py
# Python implementation of the Class QueryData
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:12:02 AM
# Original author: talve
#
#######################################################
from fetchers.ResourceDumpFetcher import ResourceDumpFetcher
from filters.SegmentsFilter import SegmentsFilter
from utils import constants


class QueryData:
    """this class is responsible for getting the menu segment.
    """

    @classmethod
    def get_query(cls, device_name, vhca_id, mem=None):
        """this method is getting the query segments by using the CoreDumpFetcher and
        filter it by removing the none menu segments.
        """
        try:
            query_kwargs = {'segment': constants.RESOURCE_DUMP_SEGMENT_TYPE_MENU,
                            'vHCAid': vhca_id, 'index1': 0, 'index2': 0, 'numOfObj1': 0, 'numOfObj2': 0, 'depth': 0, 'mem': mem, 'fast_mode': False}
            query_segments = ResourceDumpFetcher(device_name).fetch_data(**query_kwargs)
            res = SegmentsFilter.get_segments(query_segments, constants.RESOURCE_DUMP_SEGMENT_TYPE_MENU)
            try:
                menu = res[0]
            except Exception as _:
                raise Exception("Menu segment wasn't found after filtering by menu type")
            return menu
        except Exception as e:
            raise Exception("Failed to fetch query data with exception: {0}".format(e))



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# QueryCommand.py
# Python implementation of the Class QueryCommand
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:12:00 AM
# Original author: talve
#
#######################################################
from commands.ResDumpCommand import ResDumpCommand
from commands.CommandFactory import CommandFactory
from utils import constants
from validation.CapabilityValidator import CapabilityValidator
from resource_data.QueryData import QueryData
from resource_data.DataPrinter import DataPrinter
from utils import constants as cs


class QueryCommand(ResDumpCommand):
    """This class is responsible for performing the query command flow by validate,
    getting the data and print it.
    """

    def __init__(self, **kwargs):
        """QueryCommand initialization.
        """
        self.device_name = kwargs[cs.UI_ARG_DEVICE]
        self.vHCAid = kwargs[cs.UI_ARG_VHCAID]
        self.mem = kwargs[cs.UI_ARG_MEM]
        self.data = None

    def get_data(self):
        """call the QueryData for getting the menu data.
        """
        self.data = QueryData.get_query(self.device_name, self.vHCAid, self.mem)

    def print_data(self):
        """call the data printer with the right configuration for print the menu to screen.
        """
        DataPrinter.print_query_data(self.data)

    def validate(self):
        """call the capability validator and check if the core dump supported by the FW.
        """

        if CapabilityValidator.validate():
            return True
        else:
            raise Exception("Resource Dump register is not supported by the FW")


CommandFactory.register(constants.RESOURCE_DUMP_COMMAND_TYPE_QUERY, QueryCommand)



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# ResourceDumpFetcher.py
# Python implementation of the Class CoreDumpFetcher
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:12:01 AM
# Original author: talve
#
#######################################################

from codecs import utf_16_be_decode
from resource_data.DataPrinter import DataPrinter
from segments.SegmentCreator import SegmentCreator
from utils import constants
from utils import constants as cs
import sys
import os
import math
import struct
import errno
import ctypes
import platform

sys.path.append(os.path.join("mtcr_py"))
sys.path.append(os.path.join("reg_access"))

import mtcr  # noqa
import regaccess  # noqa


class MKEY_PARAMS_ST(ctypes.Structure):
    _fields_ = [("lkey", ctypes.c_uint32),
                ("umem_addr", ctypes.c_uint64)]


class EXTRACTBITS_PARAMS_ST(ctypes.Structure):
    _fields_ = [("val", ctypes.c_uint32)]


class ResourceDumpFetcher:
    """this class is responsible for getting all the segments of a the required dump.
    """
    # very big number that represent the inf number 2^32 - 1 (we will not reach that number)
    INF_DEPTH = 4294967295

    def __init__(self, device_name, bin_file=None):

        self._bin_file_h = None
        self._sequence_number = ResourceDumpFetcher._sequence_incrementor()
        self._device_name = device_name
        self._bin_file = bin_file
        self._start_seq_number = 0
        try:
            mst_device = mtcr.MstDevice(self._device_name)
            if mst_device.is_cable() or mst_device.is_linkx():
                raise Exception("Device is not supported")
            self.reg_access_obj = regaccess.RegAccess(mst_device)
        except Exception as e:
            raise Exception("{0}".format(e))

        if self._bin_file:
            try:
                self._bin_file_h = open(self._bin_file, "wb")
            except Exception as e:
                raise Exception("Failed to create binary file, {0}".format(e))

    def __del__(self):
        if self._bin_file_h:
            self._bin_file_h.close()

    @staticmethod
    def _create_segments(segments_data):
        """convert the data into a list of segments by calling the SegmentCreator.
        """
        return SegmentCreator().create(segments_data)

    def fetch_data(self, **kwargs):
        """this method fetch the segments of the required dump by:
           1. read the core dump register from the reg access while more dump bit is
        equal to  "1" and the sequence number is valid.
           2. use SegmentsCreator for converting it to a list of segments.
           3. iterate stage 1 and 2 according the depth parameter and in case reference
        segments found.
        """
        # read the inline data from the resource dump register and split it to segments list
        self._start_seq_number = 0
        inline_data = self._retrieve_resource_dump_inline_data(kwargs[cs.UI_ARG_SEGMENT], **kwargs)
        if kwargs["fast_mode"]:
            return []

        segments_list = self._create_segments(inline_data)

        segments_list_last_position = 0
        # go over oll the segments from the last position till the and of the segments list
        # and repeat it according the depth parameter
        if kwargs[cs.UI_ARG_DEPTH] == "inf":
            depth = ResourceDumpFetcher.INF_DEPTH
        else:
            depth = int(kwargs[cs.UI_ARG_DEPTH]) if kwargs[cs.UI_ARG_DEPTH] else 0
        for i in range(depth):
            inner_inline_data = []
            for seg in segments_list[segments_list_last_position:]:
                if seg.get_type() == constants.RESOURCE_DUMP_SEGMENT_TYPE_REFERENCE:
                    inner_inline_data.extend(
                        self._retrieve_resource_dump_inline_data(seg.reference_type, index1=seg.index1,
                                                                 index2=seg.index2, numOfObj1=seg.num_of_obj1,
                                                                 numOfObj2=seg.num_of_obj2, vHCAid=kwargs["vHCAid"], mem=kwargs["mem"], fast_mode=kwargs["fast_mode"]))
            segments_list_last_position = len(segments_list)
            segments_list.extend(self._create_segments(inner_inline_data))

            # relevant for inf mode, nothing to show, we can stop search for ref segments
            # because if refs found we expect that the segments list extend to do something
            if segments_list_last_position == len(segments_list):
                break

        return segments_list

    def _validate_sequence_number(self, current_seq_number):
        """validate that the sequence number was incremented.
        """
        return current_seq_number == next(self._sequence_number)

    @staticmethod
    def _sequence_incrementor():
        cnt = 0
        while True:
            yield cnt % 16
            cnt += 1

    def initSourceDump(self):
        source_dump_lib = None
        try:
            mft_py_dir = os.path.dirname(os.path.dirname(__file__))
            source_dump_lib = ctypes.CDLL(os.path.join(mft_py_dir, 'resourcedump_lib/cresourcedump.so'), use_errno=True)
        except OSError:
            raise Exception("resourcedump_lib/cresourcedump.so loading failed")
        return source_dump_lib

    def _retrieve_resource_dump_inline_data(self, segment_type, **kwargs):
        """call the resource dump access register and retrieve the inline data
        till more dump is '0'
        """
        global SCAPY_SUPPORT
        segment = int(segment_type, 16)
        seg_number = self._start_seq_number
        more_dump = 0
        vhca_id_valid = 0
        index1 = int(kwargs["index1"]) if kwargs["index1"] else 0
        index2 = int(kwargs["index2"]) if kwargs["index2"] else 0
        num_of_obj_1 = int(kwargs["numOfObj1"]) if kwargs["numOfObj1"] else 0
        num_of_obj_2 = int(kwargs["numOfObj2"]) if kwargs["numOfObj2"] else 0
        device_opaque = 0
        inline_data = []
        call_res_dump = True
        inline_dump = 1
        mkey = 0
        buf_size = 0
        address = 0
        params_st = ctypes.pointer(MKEY_PARAMS_ST())
        extract_params_st = ctypes.pointer(EXTRACTBITS_PARAMS_ST())

        if kwargs["vHCAid"] is None:
            vhca_id = 0
        else:
            vhca_id = int(kwargs["vHCAid"], 0)
            vhca_id_valid = 1

        if kwargs["mem"] is not None:
            source_dump_lib = None
            try:
                if(platform.system() != 'Linux'):
                    raise Exception("This feature supported only on Linux")
                source_dump_lib = self.initSourceDump()
                gen_lkey_ret = source_dump_lib.generate_lkey(ctypes.c_char_p(bytes(kwargs["mem"], "utf-8")), params_st)
                if gen_lkey_ret != 0:
                    raise Exception("Check the driver is up and the device's rdma name is correct")
                mkey = params_st.contents.lkey
            except Exception as e:
                raise(e)

            inline_dump = 0
            buf_size = cs.MKEY_BUF_SIZE
            address = params_st.contents.umem_addr

        while call_res_dump:

            try:
                results = self.reg_access_obj.sendResDump(segment,  # "segment_type"
                                                          seg_number,  # "seq_num"  * need check
                                                          inline_dump,  # "inline_dump"
                                                          more_dump,  # "more_dump" *
                                                          vhca_id,  # "vHCAid"
                                                          vhca_id_valid,  # "vHCAid_valid"
                                                          index1,  # "index_1" *
                                                          index2,  # "index_2" *
                                                          num_of_obj_2,  # "num_of_obj_2" *
                                                          num_of_obj_1,  # "num_of_obj_1" *
                                                          device_opaque,  # "device_opaque" *
                                                          mkey,  # "mkey"
                                                          buf_size,  # "size"
                                                          address)  # "address"
            except Exception as _:
                if _.args[0] != '':
                    raise Exception(_.args[0] + "\n Check the driver is up and the device's rdma name is correct.")

            more_dump = results["more_dump"]
            vhca_id = results["vhca_id"]
            index1 = results["index_1"]
            index2 = results["index_2"]
            num_of_obj_1 = results["num_of_obj_1"]
            num_of_obj_2 = results["num_of_obj_2"]
            device_opaque = results["device_opaque"]
            size = int(math.ceil(results["size"] / 4))
            bin_data = None

            if kwargs["mem"] is not None:
                if not kwargs["fast_mode"] or segment_type == cs.RESOURCE_DUMP_SEGMENT_TYPE_MENU:
                    # Read all dwords and change endianness
                    for i in range(size):
                        ret = source_dump_lib.bit_extracted_32(ctypes.c_uint64(address), i, extract_params_st)
                        if(ret == 0):
                            inline_data.append(extract_params_st.contents.val)
                        else:
                            raise Exception("Extracting bits from c code failed!")
                else:
                    bin_data = bytes(results["size"])
                    ret = source_dump_lib.extrac_all(ctypes.c_uint64(address), ctypes.c_int32(results["size"]), bin_data)
            else:
                if not kwargs["fast_mode"] or segment_type == cs.RESOURCE_DUMP_SEGMENT_TYPE_MENU:
                    inline_data.extend(results["inline_data"][:size])
                else:
                    bin_data = struct.pack(">{}I".format(size), *results["inline_data"][:size])
            if bin_data:
                DataPrinter.append_binary_data_to_file(bin_data, self._bin_file_h)

            if not self._validate_sequence_number(seg_number):
                raise Exception("E - wrong sequence number while calling resource dump register with seq num = {0}"
                                .format(seg_number))
            seg_number = results["seq_num"]
            self._start_seq_number = seg_number
            if more_dump == 0:
                call_res_dump = False

        return inline_data



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# AdbParser.py
# Python implementation of the Class AdbParser
# Created on:      23-Mar-2020 8:00:00 AM
# Original author: talve
#
#######################################################
import xml.etree.ElementTree as ET
import utils.constants as CONST
import ast


class AdbParser:
    """This class is responsible for loading and parsing the adb file.
    """
    def __init__(self, adb_path):
        try:
            # build a dictionary with all the adb nodes.
            self.adb_dict = ET.parse(adb_path)
            self._build_xml_elements_dict()
            self._build_definitions()
            # build a dictionary with only the adb nodes that contain segment id field.
            self.segment_id_nodes_dict = {}
            self._build_nodes_with_seg_id()
            # fix the offset since subnode should contain the offset with the addition of his parent.
            self._fix_nodes_offset(self.segment_id_nodes_dict)
            # update the union dict in the layout item if it has union_selector attribute
            self._update_union_dict()
            self._update_layout_items_paths()
            self._update_fields_with_conditions()
        except Exception as _:
            raise Exception("Failed to parse the ADB file")

    def _update_union(self, layout_item):
        """This method go over all the subitems and perform union dict updating
            to all items recursively.
        """
        if layout_item.uSelector and layout_item.uSelector.selector_full_path:
            selector_field, full_path, offset, size = self._get_explicit_field_details(layout_item.parent, layout_item.uSelector.selector_full_path)
            layout_item.uSelector.set_union_properties(self._get_enum_dict_from_field(selector_field), full_path, offset, size)
        for item in layout_item.subItems:
            self._update_union(item)

    def _update_union_dict(self):
        """This method go over all the segments in the dictionary and
        send them to union dict update.
        """
        for seg in self.segment_id_nodes_dict.values():
            self._update_union(seg)

    def check_if_exists(self, operand, item, seg):
        founded_item = None
        if('.' not in operand or "$(parent)" in operand):
            if("$(parent)" in operand):
                operand = ''.join(operand.split("$(parent)."))
            for item_it in item.parent.subItems:
                if(item_it.name == operand):
                    return item_it
        if("$(segment)" in operand or '.' in operand):
            if("$(segment)" in operand):
                operand = operand.replace("$(segment)", seg.name)
            for item_it in seg.subItems:
                if(item_it.full_path == operand):
                    return item_it
                founded_item = self.check_if_exists(operand, item, item_it)

        return founded_item

    def update_fields_elements(self, item, parent, seg):
        """This method go over all the fields in parentment and update the condtion field"""
        if(item.condition and item.parent.is_conditional):
            for operand in item.condition.operands:
                founded_item = self.check_if_exists(operand, item, seg)
                if(founded_item):
                    item.condition.var_details[founded_item.name] = ConditionVariable()
                    item.condition.var_details[founded_item.name].is_found = True
                    item.condition.var_details[founded_item.name].size = founded_item.size
                    item.condition.var_details[founded_item.name].offset = founded_item.offset
                    item.condition.var_details[founded_item.name].substring = operand
        for item_it in item.subItems:
            self.update_fields_elements(item_it, item, seg)

    def _update_fields_with_conditions(self):
        """This method go over all the segments and update fields with conditions."""

        for seg in self.segment_id_nodes_dict.values():
            for item in seg.subItems:
                self.update_fields_elements(item, seg, seg)

    def _update_layout_items_paths(self):
        for seg in self.segment_id_nodes_dict.values():
            for item in seg.subItems:
                item.full_path = seg.name + '.' + item.name
                self.update_fields_paths(item, seg)

    def update_fields_paths(self, item, parent):
        """This method go over all the fields in segment and update the path field"""

        for item_it in parent.subItems:
            if(item_it.name not in item_it.full_path and parent.name):
                item_it.full_path += parent.name + "." + item_it.name
                if(item_it.parent.full_path):
                    item_it.full_path = item_it.parent.full_path + '.' + item_it.name
            self.update_fields_paths(item_it, item)

    def _create_enum_dict(self, enum_lst): # self, enum_lst: list
        """This function gets a list of enums and export them to a dictionary.
        if the enums_lst is empty, return None.
        """
        enums_dict = {}
        if len(enum_lst) == 0:
            return None

        for enum in enum_lst:
            enum = enum.split('=')
            if len(enum) != 2:
                raise Exception("Error - Failed to parse enum attribute {0}".format(str(enum)))
            enums_dict[int(enum[1], 0)] = enum[0]
        return enums_dict

    def _get_enum_dict_from_field(self, selector_field): # self, selector_field: ET.Element
        """This method gets field that contains enums for union_selector and build a dictionary with
        it's values and enums.
        """
        # in case the enum was already parsed - we should always get in this if
        if selector_field.enum_dict:
            return selector_field.enum_dict

        if selector_field.attrs['enum'] == '':
            raise Exception("Error - enum attribute for field {0} is empty!".format(selector_field['name']))

        enum_lst = selector_field.attrs['enum'].split(',')
        selector_dict = self._create_enum_dict(enum_lst)
        if selector_dict is None:
            raise Exception("Error - could not create the enums dictionary for field {0}".format(selector_field['name']))

        return selector_dict

    def _get_explicit_field_details(self, layout_item, relative_path): # self, layout_item, relative_path: str
        """This method gets a node's name and a relative path and returns the full path to the field, the
        explicit field to the enum related to the union selector and it's offset and size.
        """
        path = relative_path.split('.')
        if len(path) == 0:
            raise Exception("Error - wrong relative path {0} for node {1}".format(layout_item.name, relative_path))
        full_path = path
        if path[0] in CONST.PARENT_LST:
            current_node = layout_item
            full_path[0] = layout_item.name
        else:
            current_node = self._retrieve_layout_item_by_name(path[0])
        try:
            if current_node is None:
                raise Exception("Failed to find node: {0}".format(current_node.name))
            for child in path[1:]:
                for item in current_node.subItems:
                    if item.name == child:
                        current_node = item
                        break
                if current_node is None:
                    raise Exception("Error - Failed to find field {0} in node {1}, wrong path {2}".format(child, current_node.name, relative_path))
            return current_node, full_path, current_node.offset, current_node.size
        except Exception as _:
            raise Exception("Failed to get the explicit field")

    def _build_xml_elements_dict(self):
        self._node_xml_elements = {}
        for xml_element in self.adb_dict.iter('node'):
            self._node_xml_elements[xml_element.attrib["name"]] = xml_element

    def _build_definitions(self):
        # Create a list for 'inst_ifdef' (C like 'define' feature)
        self.ifdef_list = []
        for xml_element in self.adb_dict.iter('config'):
            if 'define' in xml_element.attrib and "=" not in xml_element.attrib["define"]:
                self.ifdef_list.append(xml_element.attrib["define"])

        # Create a dict for 'inst_if' (C like 'define' feature)
        self.if_dict = {}
        for xml_element in self.adb_dict.iter('config'):
            if 'define' in xml_element.attrib and "=" in xml_element.attrib["define"]:
                define_name, define_value = xml_element.attrib["define"].split("=")
                self.if_dict[define_name] = define_value

    def _check_condition(self, left_operand, right_operand, condition):
        """
        This method do the right logic according the given condition
        """
        if condition == ' EQ ':
            return left_operand == right_operand
        elif condition == ' LESS ':
            return left_operand < right_operand
        elif condition == ' LESS_EQ ':
            return left_operand <= right_operand
        elif condition == ' GREAT_EQ ':
            return left_operand >= right_operand
        elif condition == ' GREAT ':
            return left_operand > right_operand

        raise Exception("unsupported condition {0}".format(condition))

    def _get_condition_str(self, expression):
        """
        This method return the right condition exist in the expression
        """
        if ' EQ ' in expression:
            return ' EQ '
        elif ' LESS ' in expression:
            return ' LESS '
        elif ' LESS_EQ ' in expression:
            return ' LESS_EQ '
        elif ' GREAT_EQ ' in expression:
            return ' GREAT_EQ '
        elif ' GREAT ' in expression:
            return ' GREAT '

        raise Exception("No condition found in expression {0}".format(expression))

    def _check_single_expression(self, expression):
        """
        The method check the condition type and perform the
        logic accordingly.
        """
        condition = self._get_condition_str(expression)
        define_name, define_value = expression.split(condition)
        define_name = define_name.strip()
        define_value = define_value.strip()

        # if operand not exist in the if dictionary its a value that we need to convert from str to int
        left_operand = int(self.if_dict[define_name]) if define_name in self.if_dict else int(define_name)
        right_operand = int(self.if_dict[define_value]) if define_value in self.if_dict else int(define_value)
        return self._check_condition(left_operand, right_operand, condition)

    def _check_expressions(self, inst_if_attrib):
        """
        The method checks an expression and return True/False.
        Note that the method support only OR(s) or AND(s) but not together.
        """
        # assumption using 'OR' or 'AND' but not together
        expressions = []
        if ' OR ' in inst_if_attrib:
            expressions = inst_if_attrib.split(' OR ')
            # check if one of the expressions is True return True
            return any([self._check_single_expression(expression) for expression in expressions])
        elif ' AND ' in inst_if_attrib:
            expressions = inst_if_attrib.split(' AND ')
            # check if all of the expressions is True return True
            return all([self._check_single_expression(expression) for expression in expressions])

        # if only one expression
        return self._check_single_expression(inst_if_attrib)

    def _fix_nodes_offset(self, nodes_dict):
        """This method go over all the nodes in the dictionary and
        send them to offset fixing.
        """
        for node in nodes_dict.values():
            self._fix_node(node, node.offset)

    def _fix_node(self, node, offset):
        """This method go over all the subnodes and perform offset addition
        to all nodes recursively.
        """
        for sub_node in node.subItems:
            sub_node.offset += offset
            self._fix_node(sub_node, sub_node.offset)

    def _build_nodes_with_seg_id(self):
        """This method go over all the nodes in the adb dictionary and build a dictionary with the nodes
        that contain the "segment_id' field.
        """
        for node in self.adb_dict.iter('node'):
            if "segment_id" in node.attrib:
                adb_layout_item = self._node_to_AdbLayoutItem(node)
                self.segment_id_nodes_dict[node.attrib["segment_id"]] = adb_layout_item
                if("is_conditional" in node.attrib):
                    if(node.attrib["is_conditional"] == '1'):
                        self.segment_id_nodes_dict[node.attrib["segment_id"]].is_conditional = True

    def _retrieve_node_by_name(self, node_name): # self, node_name: str
        """This method return the node with the
        wanted name, if not found return None.
        """
        if node_name in self._node_xml_elements:
            return self._node_xml_elements[node_name]
        else:  # Missing-nodes feature
            return None

    def _retrieve_layout_item_by_name(self, name):
        """This method return the layout item with the wanted name,
        if it wasn't found returns None
        """
        for seg in self.segment_id_nodes_dict.values():
            if seg.name == name:
                return seg
            for subItem in seg.subItems:
                if subItem.name == name:
                    return subItem
        return None

    def _build_subitems(self, node, parent): # self, node: ET.Element, parent
        """This method build the subitems of the specific node.
        :param node: always field element except from the root which is node.
        """
        sub_items = []
        counter = 0
        last_elem_attr = [elem for elem in node.iter('field')][-1].attrib
        last_elem_flag = False

        for item in node:
            # Skip on elements that are excluded in the configuration ('inst_ifdef')
            if 'inst_ifdef' in item.attrib:
                if item.attrib['inst_ifdef'] not in self.ifdef_list:
                    continue
            # Skip on elements that are excluded in the configuration ('inst_if')
            if 'inst_if' in item.attrib:
                if not self._check_expressions(item.attrib['inst_if']):
                    continue
            if(last_elem_attr == item.attrib):
                last_elem_flag = True
            if "low_bound" in item.attrib and "high_bound" in item.attrib:
                counter += self._extract_array_to_list(node, counter, item, sub_items, parent, last_elem_flag)
            else:
                adb_layout_item = self._node_to_AdbLayoutItem(item, last_elem_flag)
                if adb_layout_item:
                    adb_layout_item.parent = parent
                    adb_layout_item.attrs = item.attrib
                    if 'union_selector' in item.attrib:
                        adb_layout_item.uSelector = AdbUnionSelector()
                        adb_layout_item.uSelector.selector_full_path = item.attrib['union_selector']
                    if 'enum' in item.attrib:
                        if item.attrib['enum'] == '':
                            enums_dict = None
                        else:
                            enums_lst = item.attrib['enum'].split(',')
                            enums_dict = self._create_enum_dict(enums_lst)
                            if enums_dict is None:
                                raise Exception("Error - could not create the enums dictionary for node {0}".format(item.attrib['name']))
                        adb_layout_item.enum_dict = enums_dict
                    sub_items.insert(counter, adb_layout_item)
                    counter += 1
        return sub_items

    def _extract_array_to_list(self, node, index, item, subitems_list, parent, last_elem_flag=False): # self, node: ET.Element, index, item, subitems_list, parent, last_elem_flag=False
        """This method build insert array items to a list form a specific index.
        :param node: always field element except from the root which is node.
        :param parent: layout item representing the parent of the array.
        """
        name = item.attrib["name"]
        start_index = int(item.attrib["low_bound"], 10)
        if item.attrib["high_bound"] == "VARIABLE":
            end_index = 1
        else:
            end_index = int(item.attrib["high_bound"], 10) + 1

        offset = self._parse_node_size(item.attrib["offset"])
        size = self._parse_node_size(item.attrib["size"])

        calculated_size = int(size / (end_index - start_index))
        current_offset = offset

        enum_dict = None
        if 'index_enum' in item.attrib:
            if item.attrib['index_enum'] == '':
                enum_dict = None
            else:
                enums_lst = item.attrib['index_enum'].split(',')
                enum_dict = self._create_enum_dict(enums_lst)
                if enum_dict is None:
                    raise Exception("Error - could not parse the 'index_enum' attribute for node .".format(name))

        for i in range(start_index, end_index):
            adb_layout_item = AdbLayoutItem()
            # adb_layout_item.attrs = node.attrib
            adb_layout_item.parent = parent
            adb_layout_item.parent.size = self._parse_node_size(node.attrib["size"])
            adb_layout_item.nodeDesc = self._node_to_node_desc(item)
            if enum_dict is None or i not in enum_dict:
                adb_layout_item.name = name + "[" + str(i) + "]"
            else:
                adb_layout_item.name = name + "[" + enum_dict[i] + "]"
            adb_layout_item.size = calculated_size
            adb_layout_item.offset = current_offset
            if(last_elem_flag):
                adb_layout_item.last_elem = True
            if(item.attrib['high_bound'] == 'VARIABLE' and last_elem_flag and (parent.last_elem or 'segment_id' in node.attrib)):
                adb_layout_item.is_unlimited_array = True
            try:
                if "subnode" in item.attrib:
                    sub_node = self._retrieve_node_by_name(item.attrib["subnode"])
                    if sub_node:
                        if "attr_is_union" in sub_node.attrib:
                            adb_layout_item.nodeDesc.isUnion = self._parse_union(sub_node.attrib["attr_is_union"])
                        adb_layout_item.subItems = self._build_subitems(sub_node, parent)  # List of the child items (for nodes only)
            except Exception as _:
                raise Exception("Failed to extract array to list from node {0}".format(node['name']))
            subitems_list.insert(index, adb_layout_item)
            current_offset += calculated_size
            index += 1
        return index

    def _node_to_AdbLayoutItem(self, node, last_elem_flag=False): # self, node: ET.Element, last_elem_flag=False
        """This method build adb layout field from a given node.
        :param node: always field element except from the root which is node.
        """
        adb_layout_item = AdbLayoutItem()
        # adb_layout_item.fieldDesc = None  # Reference to "AdbFieldDesc" object, always != None
        # adb_layout_item.parent = None  # Reference to parent "AdbLayoutItem" object, always != None (expect of the root)
        adb_layout_item.name = node.attrib["name"]
        adb_layout_item.attrs = node.attrib
        if(last_elem_flag):
            adb_layout_item.last_elem = True
        if("condition" in node.attrib and node.attrib["condition"]):
            for conf in self.if_dict:
                for op in ConditionParser.operators:
                    if(conf in node.attrib["condition"] and (op + conf or conf + op) in node.attrib["condition"] and
                       not ("." + conf or conf + ".") in node.attrib["condition"]):
                        node.attrib["condition"] = node.attrib["condition"].replace(conf, self.if_dict[conf])
            adb_layout_item.condition = ConditionParser(node.attrib["condition"])
        if("is_conditional" in node.attrib):
            if(node.attrib["is_conditional"] == '1'):
                adb_layout_item.is_conditional = True

        adb_layout_item.nodeDesc = self._node_to_node_desc(node)  # For leafs must be None
        try:
            if "subnode" in node.attrib:
                sub_node = self._retrieve_node_by_name(node.attrib["subnode"])
                if "attr_is_union" in sub_node.attrib:
                    adb_layout_item.nodeDesc.isUnion = self._parse_union(sub_node.attrib["attr_is_union"])
                adb_layout_item.subItems = self._build_subitems(sub_node, adb_layout_item)  # List of the child items (for nodes only)
            else:
                adb_layout_item.subItems = self._build_subitems(node, adb_layout_item)
        except Exception as _:
            raise Exception("Failed to create a valid layout from node '{0}'".format(node['name']))

        if "offset" in node.attrib:
            adb_layout_item.offset = self._parse_node_size(node.attrib["offset"])

        adb_layout_item.size = self._parse_node_size(node.attrib["size"])
        adb_layout_item.attrs = {}  # Attributes after evaluations and array expanding
        adb_layout_item.arrayIdx = -1  # in case of being part of an array this will hold the array index, -1 means not part of array
        # adb_layout_item.vars = {}  # all variable relevant to this item after evaluation
        return adb_layout_item

    def _node_to_node_desc(self, node): # self, node: ET.Element
        """This method build adb desc field from a given node.
        """
        node_description = AdbNodeDesc()
        #node_description.name = ""
        # node_description.size = 0  # Size in bits
        if "attr_is_union" in node.attrib:
            node_description.isUnion = self._parse_union(node.attrib["attr_is_union"])
        #node_description.desc = ""
        # node_description.fields = []  # List of "AdbFieldDesc" objects
        # node_description.attrs = {}  # Dictionary of attributes: key, value
        # defined in
        #node_description.fileName = ""
        #node_description.lineNumber = -1
        return node_description

    def _parse_union(self, union_str):
        """This method parse the 'is union' attribute.
        """
        if union_str == "1":
            return True
        return False

    def _parse_node_size(self, size_str):
        """This method return the size of the 'size' attribute in bits.
        """
        hex_size = 0
        bit_size = 0
        size_parts = size_str.split(".")

        if len(size_parts) == 1:
            size_parts = size_parts[0], "0"
        if size_parts[0] == "":
            size_parts[0] = "0"

        hex_size = int(size_parts[0], 16) * 8
        bit_size = int(size_parts[1], 10)

        return hex_size + bit_size

####################################################################


class ConditionVariable(object):
    def __init__(self):
        self.is_found = False
        self.value = 0
        self.offset = 0
        self.size = 0
        self.substring = ""


class ConditionParser(object):
    ops = {'~': 'not ', 'AND': 'and', '|': 'or'}
    operators = ['~', '|', 'AND', '==', '!=', '(', ')']

    def __init__(self, condition):
        self.operands = set()
        self.parsed_str = self.parse_str(condition)
        self.var_details = {}
        self.result = False

    def parse_str(self, str):
        lst = str.split(" ")
        for elem in lst:
            tmp_elem = elem
            for op in ConditionParser.operators:
                if op in elem:
                    if op == '==' or op == '!=':
                        tmp_elem = elem.split(op)[0]
                    tmp_elem = tmp_elem.replace(op, "")
            if("segment" in tmp_elem):
                tmp_elem = tmp_elem.replace("segment", "(segment)")
            if("parent" in tmp_elem):
                tmp_elem = tmp_elem.replace("parent", "(parent)")
            self.operands.add(tmp_elem)
        if('' in self.operands):
            self.operands.remove('')
        for key in ConditionParser.ops:
            if key in str:
                str = str.replace(key, ConditionParser.ops[key])
        return str

    def eval_expr(self):
        try:
            tree = ast.parse(self.parsed_str, mode='eval')
        except SyntaxError:
            return    # not a Python expression
        if not all(isinstance(node, (ast.Expression,
                                     ast.UnaryOp, ast.unaryop,
                                     ast.BinOp, ast.operator,
                                     ast.BoolOp, ast.Num, ast.And, ast.Or,
                                     ast.Compare, ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE)) for node in ast.walk(tree)):
            return    # not a mathematical expression (numbers and op)
        return eval(compile(tree, filename='', mode='eval'))


class AdbLayoutItem(object):
    """ Represents a Layout node/leaf instance """

    ##########################
    def __init__(self):
        self.fieldDesc = None        # Reference to "AdbFieldDesc" object, always != None
        self.nodeDesc = None         # For leafs must be None
        self.parent = None           # Reference to parent "AdbLayoutItem" object, always != None (expect of the root)
        self.name = ""               # The instance name
        self.subItems = []           # List of the child items (for nodes only)
        self.offset = 0              # Global offset (relative to the 0 address)
        self.size = 0                # Element size in bits
        self.attrs = {}              # Attributes after evaluations and array expanding
        self.arrayIdx = -1           # in case of being part of an array this will hold the array index, -1 means not part of array
        self.vars = {}               # all variable relevant to this item after evaluation
        self.uSelector = None        # data structure that represent the union-selector properties
        self.enum_dict = None        # in case of being an enum element, holds the key-value of the enum
        self.condition = None
        self.full_path = ""
        self.is_conditional = False
        self.last_elem = False
        self.is_unlimited_array = False


####################################################################
class AdbFieldDesc(object):
    """ Represents an ADABE field descriptor, objects of this type are immutable """

    ##########################
    def __init__(self):
        self.name = ""
        self.size = 0                    # Size in bits (the whole array size)
        self.offset = None                 # Offset in bits (relative to the node start addr)
        self.desc = ""
        self.lowBound = 0
        self.highBound = 0
        self.unlimitedArr = False
        self.definedAsArr = False
        self.subNode = None                # Subnode name as string
        self.attrs = {}                  # Dictionary of attributes: key, value
        self.reserved = False
        self.condition = ""


####################################################################
class AdbNodeDesc(object):
    """ Represents an ADABE node descriptor, objects of this type are immutable """

    ##########################
    def __init__(self):
        self.name = ""
        self.size = 0                 # Size in bits
        self.isUnion = False
        self.desc = ""
        self.fields = []                # List of "AdbFieldDesc" objects
        self.attrs = {}                # Dictionary of attributes: key, value

        # defined in
        self.fileName = ""
        self.lineNumber = -1


####################################################################
class AdbUnionSelector(object):
    """ Represents an ADABE union-selector descriptor, objects of this type are immutable """

    ##########################
    def __init__(self):
        self.dict = None               # mapping between the value and the enum, None if empty
        self.full_path = None          # full path to the selector field, None if empty
        self.offset = None             # absolute offset from the begining of the segment until the selector field (by layout)
        self.size = None               # size of the selector itself

    def set_union_properties(self, union_dict, full_path, offset, size):
        self.dict = union_dict
        self.full_path = full_path
        self.offset = offset
        self.size = size



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# TerminateSegment.py
# Python implementation of the Class TerminateSegment
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:11:57 AM
# Original author: talve
#
#######################################################
from segments.Segment import Segment
from segments.SegmentFactory import SegmentFactory
from utils import constants


class TerminateSegment(Segment):
    """this class is responsible for holding Terminate segment data.
    """
    def __init__(self, data):
        """initialize the class by setting the class data.
        """
        self.raw_data = data

    def get_data(self):
        """get the general segment data.
        """
        return self.raw_data

    def get_type(self):
        """get the general segment type.
        """
        return constants.RESOURCE_DUMP_SEGMENT_TYPE_TERMINATE


SegmentFactory.register(constants.RESOURCE_DUMP_SEGMENT_TYPE_TERMINATE, TerminateSegment)



# Copyright (C) Jan 2020 Mellanox Technologies Ltd. All rights reserved.
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# This software is available to you under a choice of one of two
# licenses.  You may choose to be licensed under the terms of the GNU
# General Public License (GPL) Version 2, available from the file
# COPYING in the main directory of this source tree, or the
# OpenIB.org BSD license below:
#
#     Redistribution and use in source and binary forms, with or
#     without modification, are permitted provided that the following
#     conditions are met:
#
#      - Redistributions of source code must retain the above
#        copyright notice, this list of conditions and the following
#        disclaimer.
#
#      - Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials
#        provided with the distribution.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --


#######################################################
#
# DataPrinter.py
# Python implementation of the Class DataPrinter
# Generated by Enterprise Architect
# Created on:      14-Aug-2019 10:12:04 AM
# Original author: talve
#
#######################################################

from __future__ import print_function


class DataPrinter:
    """this class is responsible for print the given segments to the right output.
    """
    NUM_OF_DW_IN_ROW = 4

    @classmethod
    def print_query_data(cls, data):
        """this method print the user menu from the menu segment data
        """
        if data:
            for record in data.get_printable_records():
                for el in record:
                    print(el)

    @classmethod
    def print_dump_data(cls, data, file):
        """the method gets a list of segments and arguments if print to bin file or output
        to screen.
        """
        if data:
            if file:
                DataPrinter._write_to_file(data, file)
            else:
                print("Found {0} segments:".format(len(data)))
                for seg in data:
                    rows = 0
                    print("-" * 43)
                    print("Segment Type: {}".format(seg.get_type()))
                    print("Segment Size: {} Bytes".format(len(seg.get_data()) * 4))
                    print("Segment Data:")
                    for dw in seg.get_data():
                        print('0x{0:0{1}X} '.format(dw, 8), end="")
                        rows += 1
                        if rows == DataPrinter.NUM_OF_DW_IN_ROW:
                            print("")
                            rows = 0
                    if 0 < rows < DataPrinter.NUM_OF_DW_IN_ROW:
                        print("")
                print("-" * 43)

    @classmethod
    def _write_to_file(cls, data, path):
        with open(path, "wb") as binary_file:
            for seg in data:
                for dw in seg.get_data():
                    four_bytes = dw.to_bytes(4, byteorder='big', signed=False)
                    binary_file.write(four_bytes)

        print("write to file: ", path)

    @classmethod
    def append_binary_data_to_file(cls, data, file_h):
        file_h.write(data)
